// DOM Utils
import getElementClientRect from './dom-utils/getRectRelativeToOffsetParent';
import listScrollParents from './dom-utils/listScrollParents';
import getWindow from './dom-utils/getWindow';
import addClientRectMargins from './dom-utils/addClientRectMargins'; // Pure Utils

import unwrapJqueryElement from './utils/unwrapJqueryElement';
import orderModifiers from './utils/orderModifiers';
import expandToHashMap from './utils/expandToHashMap';
import debounce from './utils/debounce';
import validateModifiers from './utils/validateModifiers';
var INVALID_ELEMENT_ERROR = 'Popper: Invalid reference or popper argument provided to Popper, they must be either a valid DOM element, virtual element, or a jQuery-wrapped DOM element.';
var INFINITE_LOOP_ERROR = 'Popper: An infinite loop in the modifiers cycle has been detected! The cycle has been interrupted to prevent a browser crash.';

var areValidElements = function areValidElements() {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }

  return !args.some(function (element) {
    return !(element && typeof element.getBoundingClientRect === 'function');
  });
};

var defaultOptionsValue = {
  placement: 'bottom',
  eventListeners: {
    scroll: true,
    resize: true
  },
  modifiers: [],
  strategy: 'absolute'
};
export function popperGenerator(generatorOptions) {
  if (generatorOptions === void 0) {
    generatorOptions = {};
  }

  var _generatorOptions = generatorOptions,
      _generatorOptions$def = _generatorOptions.defaultModifiers,
      defaultModifiers = _generatorOptions$def === void 0 ? [] : _generatorOptions$def,
      _generatorOptions$def2 = _generatorOptions.defaultOptions,
      defaultOptions = _generatorOptions$def2 === void 0 ? defaultOptionsValue : _generatorOptions$def2;
  return function createPopper(reference, popper, options) {
    if (options === void 0) {
      options = defaultOptions;
    }

    var instance = {};
    var state = {
      placement: 'bottom',
      orderedModifiers: [],
      options: defaultOptions,
      modifiersData: {}
    }; // Syncronous and forcefully executed update
    // it will always be executed even if not necessary, usually NOT needed
    // use Popper#update instead

    instance.forceUpdate = function () {
      var _state$elements = state.elements,
          referenceElement = _state$elements.reference,
          popperElement = _state$elements.popper; // Don't proceed if `reference` or `popper` are not valid elements anymore

      if (!areValidElements(referenceElement, popperElement)) {
        if (process.env.NODE_ENV !== "production") {
          console.error(INVALID_ELEMENT_ERROR);
        }

        return;
      }

      var isFixed = state.options.strategy === 'fixed'; // Get initial measurements
      // these are going to be used to compute the initial popper offsets
      // and as cache for any modifier that needs them later

      state.measures = {
        reference: getElementClientRect(referenceElement, isFixed),
        // CSS marginsc an be applied to popper elements to quickly
        // apply offsets dynamically based on some CSS selectors.
        // For this reason we include margins in this calculation.
        popper: addClientRectMargins(getElementClientRect(popperElement, isFixed), popperElement)
      }; // Modifiers have the ability to read the current Popper state, included
      // the popper offsets, and modify it to address specifc cases

      state.reset = false; // Cache the placement in cache to make it available to the modifiers
      // modifiers will modify this one (rather than the one in options)

      state.placement = state.options.placement;
      state.orderedModifiers.forEach(function (modifier) {
        return state.modifiersData[modifier.name] = Object.assign({}, modifier.data);
      });
      var __debug_loops__ = 0;

      for (var index = 0; index < state.orderedModifiers.length; index++) {
        if (process.env.NODE_ENV !== "production") {
          __debug_loops__ += 1;

          if (__debug_loops__ > 100) {
            console.error(INFINITE_LOOP_ERROR);
            break;
          }
        }

        if (state.reset === true) {
          state.reset = false;
          index = -1;
          continue;
        }

        var _state$orderedModifie = state.orderedModifiers[index],
            fn = _state$orderedModifie.fn,
            enabled = _state$orderedModifie.enabled,
            _options = _state$orderedModifie.options;

        if (enabled && typeof fn === 'function') {
          state = fn(state, _options);
        }
      }
    }; // Async and optimistically optimized update
    // it will not be executed if not necessary
    // debounced, so that it only runs at most once-per-tick


    instance.update = debounce(function () {
      return new Promise(function (resolve) {
        instance.forceUpdate();
        resolve(state);
      });
    }); // Unwrap `reference` and `popper` elements in case they are
    // wrapped by jQuery, otherwise consume them as is

    var referenceElement = unwrapJqueryElement(reference);
    var popperElement = unwrapJqueryElement(popper);
    state.elements = {
      reference: referenceElement,
      popper: popperElement
    }; // Store options into state

    state.options = Object.assign({}, defaultOptions, {}, options); // Don't proceed if `reference` or `popper` are invalid elements

    if (!areValidElements(referenceElement, popperElement)) {
      if (process.env.NODE_ENV !== "production") {
        console.error(INVALID_ELEMENT_ERROR);
      }

      return;
    }

    state.scrollParents = {
      reference: listScrollParents(referenceElement),
      popper: listScrollParents(popperElement)
    }; // Validate the provided modifiers so that the consumer will get warned
    // of one of the custom modifiers is invalid for any reason

    if (process.env.NODE_ENV !== "production") {
      validateModifiers(state.options.modifiers);
    } // Order `options.modifiers` so that the dependencies are fulfilled
    // once the modifiers are executed


    state.orderedModifiers = orderModifiers([].concat(defaultModifiers, state.options.modifiers)) // Apply user defined preferences to modifiers
    .map(function (modifier) {
      return Object.assign({}, modifier, {}, state.options.modifiers.find(function (_ref) {
        var name = _ref.name;
        return name === modifier.name;
      }));
    }); // Modifiers have the opportunity to execute some arbitrary code before
    // the first update cycle is ran, the order of execution will be the same
    // defined by the modifier dependencies directive.
    // The `onLoad` function may add or alter the options of themselves

    state.orderedModifiers.forEach(function (_ref2) {
      var onLoad = _ref2.onLoad,
          enabled = _ref2.enabled;
      return enabled && onLoad && onLoad(state);
    });

    instance.enableEventListeners = function (eventListeners) {
      var _ref3 = typeof eventListeners === 'boolean' ? expandToHashMap(eventListeners, ['scroll', 'resize']) : eventListeners,
          scroll = _ref3.scroll,
          resize = _ref3.resize;

      if (scroll) {
        var scrollParents = [].concat(state.scrollParents.reference, state.scrollParents.popper);
        scrollParents.forEach(function (scrollParent) {
          return scrollParent.addEventListener('scroll', instance.update, {
            passive: true
          });
        });
      }

      if (resize) {
        var window = getWindow(state.elements.popper);
        window.addEventListener('resize', instance.update, {
          passive: true
        });
      }
    };

    instance.destroy = function () {
      // Remove scroll event listeners
      var scrollParents = [].concat(state.scrollParents.reference, state.scrollParents.popper);
      scrollParents.forEach(function (scrollParent) {
        return scrollParent.removeEventListener('scroll', instance.update);
      }); // Remove resize event listeners

      var window = getWindow(state.elements.popper);
      window.removeEventListener('resize', instance.update); // Run `onDestroy` modifier methods

      state.orderedModifiers.forEach(function (_ref4) {
        var onDestroy = _ref4.onDestroy,
            enabled = _ref4.enabled;
        return enabled && onDestroy && onDestroy(state);
      });
    };

    instance.update().then(function () {
      // After the first update completed, enable the event listeners
      instance.enableEventListeners(state.options.eventListeners);
    });
    return instance;
  };
}
export var createPopper = popperGenerator();